#include <stdio.h>
#include <process.h>

#include <OB/Basic.h>
#include <OB/Except.h>
#include <OB/Reactor.h>
#include <OB/Net.h>
#include <OB/Util.h>

extern "C" {
#include <lua.h>
}

#include "wreactor-ob3.h"

#define WM_WINSOCK (WM_USER + 1)
#define WM_CONSOLE_INPUT (WM_USER + 2)

typedef unsigned(__stdcall* THREAD_ADDRESS)(void*);

struct ThreadState {
   HWND hwnd;
   HANDLE handle;
   char code[500];
};

static ThreadState state_;

// ----------------------------------------------------------------------
// Windows message processing function
// ----------------------------------------------------------------------

static WConsoleReactor* WinReactor = 0;

unsigned long WConsoleReactor::thread_handle_ = 0;
HWND WConsoleReactor::hWnd_ = 0;

static LRESULT CALLBACK
WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    assert(WinReactor);

    return WinReactor -> wndProc(hWnd, message, wParam, lParam);
}


// ----------------------------------------------------------------------
// Console's input handler
// ----------------------------------------------------------------------

static unsigned int stdin_handler(ThreadState* state)
{
   while (fgets(state->code,500,stdin))
   {
      PostMessage(state->hwnd,WM_CONSOLE_INPUT,NULL,(LPARAM) state);
      WaitForSingleObject(state->handle,INFINITE);
      if (state->handle)
      {
         ResetEvent(state->handle);
         printf("> ");
         fflush(stdout);
      }
      else
         break;
   }
   return 0;
}

void WConsoleReactor::Start()
{
   printf("LuaOrb Console v. 1.5\n> ");
   fflush(stdout);

   state_.hwnd = WConsoleReactor::hWnd_;
   state_.handle = CreateEvent(NULL,1,0,NULL);

   unsigned int thread_id;
   WConsoleReactor::thread_handle_ = _beginthreadex(NULL,0, (THREAD_ADDRESS)stdin_handler,
                                                    &state_, 0, &thread_id);
   if (WConsoleReactor::thread_handle_ <= 0)
   {
      WConsoleViewer::instance()->message("LuaOrb Console: Error creating input thread.");
   }
}

void WConsoleReactor::Close()
{
   HANDLE old_handle = state_.handle;
   state_.handle = 0;

   if (WConsoleReactor::thread_handle_ > 0)
   {
      if (old_handle)
         SetEvent(old_handle);
      CloseHandle((HANDLE)WConsoleReactor::thread_handle_);
      WConsoleReactor::thread_handle_ = 0;
   }

   if (old_handle)
      CloseHandle(old_handle);
}

// ----------------------------------------------------------------------
// WConsoleReactor constructor and destructor
// ----------------------------------------------------------------------

WConsoleReactor::WConsoleReactor(HINSTANCE hInstance)
    : stop_(false), handlerInfoList_(0)
{
   assert(WinReactor == 0);
   WinReactor = this;
   
   static char szAppName[] = "LuaOrbReactor";
   WNDCLASSEX  wndclass;
    
   wndclass.cbSize        = sizeof(wndclass);
   wndclass.style         = 0;
   wndclass.lpfnWndProc   = WndProc;
   wndclass.cbClsExtra    = 0;
   wndclass.cbWndExtra    = 0;
   wndclass.hInstance     = hInstance;
   wndclass.hIcon         = 0;
   wndclass.hCursor       = 0;
   wndclass.hbrBackground = 0;
   wndclass.lpszMenuName  = 0;
   wndclass.lpszClassName = szAppName;
   wndclass.hIconSm       = 0;
    
   RegisterClassEx(&wndclass);
    
   hWnd_ = CreateWindow(szAppName,
                        "The LuaOrb Reactor",
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        0,
                        0,
                        hInstance,
                        0);
}

WConsoleReactor::~WConsoleReactor()
{
   DestroyWindow(hWnd_);

   assert(WinReactor == this);
   WinReactor = 0;

   //
   // Delete all handler infos
   //
   while(handlerInfoList_)
   {
      HandlerInfo* next = handlerInfoList_ -> next;
      delete handlerInfoList_;
      handlerInfoList_ = next;
   }
}

// ----------------------------------------------------------------------
// WConsoleReactor public member implementation
// ----------------------------------------------------------------------

OBReactor* WConsoleReactor::instance(HINSTANCE hInstance)
{
   if(!instance_)
      instance_ = new WConsoleReactor(hInstance);

   return instance_;
}

void WConsoleReactor::registerHandler(OBEventHandler* handler, OBMask mask, OBHandle handle)
{
   assert(handler);

   HandlerInfo* info = handlerInfoList_;
   while(info)
   {
      if(info -> handler == handler)
      {
         //
         // Handler already in list, change notification messages
         // and update mask and handle if necessary
         //
         if(info -> mask != mask || info -> handle != handle)
         {
            WSAAsyncSelect(info -> handle, hWnd_, WM_WINSOCK, 0);

            info -> handle = handle;
            info -> mask = mask;

            long lEvent = 0;

            if (mask & OBEventRead)
               lEvent |= FD_READ | FD_ACCEPT | FD_CLOSE;

            if (mask & OBEventWrite)
               lEvent |= FD_WRITE | FD_CONNECT;

            WSAAsyncSelect(info -> handle, hWnd_, WM_WINSOCK, lEvent);
         }

         return;
      }
	
      info = info -> next;
   }

   //
   // Add new handler info
   //
   info = new HandlerInfo;
   info -> handler = handler;
   info -> mask = mask;
   info -> handle = handle;
   info -> next = handlerInfoList_;
   handlerInfoList_ = info;

   //
   // Add notification messages
   //
   long lEvent = 0;

   if (mask & OBEventRead)
      lEvent |= FD_READ | FD_ACCEPT | FD_CLOSE;
    
   if (mask & OBEventWrite)
      lEvent |= FD_WRITE | FD_CONNECT;
    
   WSAAsyncSelect(info -> handle, hWnd_, WM_WINSOCK, lEvent);
}

void WConsoleReactor::unregisterHandler(OBEventHandler* handler)
{
   assert(handler);

   HandlerInfo** p = &handlerInfoList_;
   while(*p)
   {
      HandlerInfo* info = *p;

      if(info -> handler == handler)
      {
         //
         // Handler info found, remove notification messages and
         // invalidate the handler
         //
         WSAAsyncSelect(info -> handle, hWnd_, WM_WINSOCK, 0);
         *p = info -> next;
         delete info;
         return;
      }
	
      p = &(info -> next);
   }

   assert(false);
}

void WConsoleReactor::dispatch()
{
   MSG msg;
   while(!stop_ && handlerInfoList_ && GetMessage(&msg, 0, 0, 0))
   {
      //
      // Process one Windows message
      //
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }
}

bool WConsoleReactor::dispatchOneEvent(CORBA_Long timeout)
{
   assert_nca(timeout == -1 || timeout == 0, OBNCANotImpl);

   if (stop_ || !handlerInfoList_)
      return false; // No event

   MSG msg;
   BOOL result;

   if (timeout == -1)
      result = GetMessage(&msg, 0, 0, 0);
   else if(timeout == 0)
      result = PeekMessage(&msg, 0, 0, 0, PM_REMOVE);

   if(result)
   {
      //
      // Process one Windows message
      //
      TranslateMessage(&msg);
      DispatchMessage(&msg);

      return true;
   }
   else
      return false;
}

LRESULT WConsoleReactor::wndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   if(hWnd == hWnd_)
   {
      switch (message)
      {
      case WM_WINSOCK:
      {
         HandlerInfo* info = handlerInfoList_;
         while(info)
         {
            if((WPARAM)info -> handle == wParam)
            {
	            //
            	// Calculate the event mask
               //
               OBMask mask = 0;

               switch(WSAGETSELECTEVENT(lParam))
               {
               case FD_READ:
               case FD_ACCEPT:
               case FD_CLOSE:
                  mask |= OBEventRead;
                  break;
               case FD_WRITE:
               case FD_CONNECT:
                  mask |= OBEventWrite;
                  break;
               }

               //
               // Execute the event
               //
               if(info -> mask & mask)
                  info -> handler -> handleEvent(mask);
		
               //
               // No further looping necessary, only one event can be
               // delivered through this operation
               //
               return 0;
            }

            info = info -> next;
         }
         break;
      }
      case WM_CONSOLE_INPUT:
      {
         if (lParam)
         {
            ThreadState* state = (ThreadState*) lParam;
            lua_dostring(state->code);
            if (state->handle)
               SetEvent(state->handle);
         }
         break;
      }
      }
   }

   return DefWindowProc(hWnd, message, wParam, lParam);
}

// ----------------------------------------------------------------------
// WConsoleViewer public member implementation
// ----------------------------------------------------------------------

OBMessageViewer* WConsoleViewer::instance()
{
    if(!instance_)
	instance_ = new WConsoleViewer();

    return instance_;
}

void WConsoleViewer::message(const char* msg)
{
   printf("%s\n",msg);
   fflush(stdout);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

void WConsoleReactor::Init(const char* title)
{
    HINSTANCE hInstance = GetModuleHandle(NULL);
    SetConsoleTitle(title);
    state_.hwnd = 0;
    state_.handle = 0;
    state_.code[0]= 0;

    //
    // This will choose WConsoleReactor as reactor singleton
    //
    WConsoleReactor::instance(hInstance);

    //
    // This will choose WConsoleViewer as message viewer singleton
    //
    WConsoleViewer::instance();
}

