#include <stdio.h>
#include <process.h>

#include <OB/Except.h>
#include <OB/OBWindows.h>
#include <OB/Object.h>
#include <OB/EventHandlerBase.h>

extern "C" {
#include <lua.h>
}

#include "wreactor-ob4.h"

#define WM_WINSOCK (WM_USER + 1)
#define WM_CONSOLE_INPUT (WM_USER + 2)

using namespace CORBA;
using namespace OB;

typedef unsigned(__stdcall* THREAD_ADDRESS)(void*);

struct ThreadState {
   HWND hwnd;
   HANDLE handle;
   char code[500];
};

static ThreadState state_;


//
// There may only be one WConsoleReactor per-process
//
static WConsoleReactor* instance_ = 0;

unsigned long WConsoleReactor::thread_handle_ = 0;
HWND WConsoleReactor::hWnd_ = 0;

// ----------------------------------------------------------------------
// Windows message processing function
// ----------------------------------------------------------------------

static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   assert(instance_);
   return instance_->wndProc(hWnd, message, wParam, lParam);
}


// ----------------------------------------------------------------------
// Console's input handler
// ----------------------------------------------------------------------

static unsigned int stdin_handler(ThreadState* state)
{
   while (fgets(state->code,500,stdin))
   {
      PostMessage(state->hwnd, WM_CONSOLE_INPUT, NULL, (LPARAM) state);
      WaitForSingleObject(state->handle,INFINITE);
      if (state->handle)
      {
         ResetEvent(state->handle);
         printf("> ");
         fflush(stdout);
      }
      else
         break;
   }
   return 0;
}

void WConsoleReactor::Start()
{
   printf("LuaOrb Console v. 1.5\n> ");
   fflush(stdout);

   state_.hwnd = WConsoleReactor::hWnd_;
   state_.handle = CreateEvent(NULL,1,0,NULL);

   unsigned int thread_id;
   WConsoleReactor::thread_handle_ = _beginthreadex(NULL,0,(THREAD_ADDRESS)stdin_handler,
                                                    &state_, 0, &thread_id);
   if (WConsoleReactor::thread_handle_ <= 0)
   {
      fprintf(stderr,"LuaOrb Console: Error creating input thread.\n");
   }
}

void WConsoleReactor::Close()
{
   HANDLE old_handle = state_.handle;
   state_.handle = 0;

   if (WConsoleReactor::thread_handle_ > 0)
   {
      if (old_handle)
         SetEvent(old_handle);
      CloseHandle((HANDLE)WConsoleReactor::thread_handle_);
      WConsoleReactor::thread_handle_ = 0;
   }

   if (old_handle)
      CloseHandle(old_handle);
}

// ----------------------------------------------------------------------
// WConsoleReactor private member implementation
// ----------------------------------------------------------------------

TypeMask WConsoleReactor::evalTypeMasks() const
{
   OB_SYNCHRONIZED

   //
   // Add all event handlers
   //
   TypeMask typeMask = 0;
   for (OB_STL::list<HandlerInfo>::const_iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end(); iter++)
   {
      typeMask |= (*iter).typeMask;
   }

   return typeMask;
}

// ----------------------------------------------------------------------
// WConsoleReactor constructor and destructor
// ----------------------------------------------------------------------

WConsoleReactor::WConsoleReactor(HINSTANCE hInstance)
   : interrupt_(false)
{
   assert(instance_ == 0);
   instance_ = this;
   
   static char szAppName[] = "LuaOrbReactor";
   WNDCLASSEX  wndclass;
   
   wndclass.cbSize        = sizeof(wndclass);
   wndclass.style         = 0;
   wndclass.lpfnWndProc   = WndProc;
   wndclass.cbClsExtra    = 0;
   wndclass.cbWndExtra    = 0;
   wndclass.hInstance     = hInstance;
   wndclass.hIcon         = 0;
   wndclass.hCursor       = 0;
   wndclass.hbrBackground = 0;
   wndclass.lpszMenuName  = 0;
   wndclass.lpszClassName = szAppName;
   wndclass.hIconSm       = 0;
   
   RegisterClassEx(&wndclass);
   
   hWnd_ = CreateWindow(szAppName,
                        "The LuaOrb Reactor",
                        WS_OVERLAPPEDWINDOW,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        CW_USEDEFAULT,
                        0,
                        0,
                        hInstance,
                        0);
}

WConsoleReactor::~WConsoleReactor()
{
   DestroyWindow(hWnd_);

   assert(instance_ == this);
   instance_ = 0;
}

// ----------------------------------------------------------------------
// WConsoleReactor public member implementation
// ----------------------------------------------------------------------

void WConsoleReactor::register_handler(EventHandlerBase_ptr handler, Mask mask, TypeMask typeMask, Handle handle)
{
   OB_SYNCHRONIZED

   assert(!is_nil(handler));

   for (OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end(); iter++)
   {
      HandlerInfo& info = *iter;

      if(info.handler.in() == handler)
      {
         //
         // Handler already in list, change notification messages
         // and update mask and handle if necessary
         //
         if(info.mask != mask || info.handle != handle)
         {
            info.handle = handle;
            info.mask = mask;

            long lEvent = 0;

            if (mask & EventRead)
               lEvent |= FD_READ | FD_ACCEPT | FD_CLOSE;

            if(mask & EventWrite)
               lEvent |= FD_WRITE | FD_CONNECT;

            WSAAsyncSelect(info.handle, hWnd_, WM_WINSOCK, lEvent);
         }

         info.typeMask = typeMask;
         return;
      }
   }

   //
   // Setup new handler info
   //
   HandlerInfo info;
   info.handler = EventHandlerBase::_duplicate(handler);
   info.handle = handle;
   info.mask = mask;
   info.typeMask = typeMask;

   //
   // Add notification messages
   //
   long lEvent = 0;

   if (mask & EventRead)
      lEvent |= FD_READ | FD_ACCEPT | FD_CLOSE;
   
   if (mask & EventWrite)
      lEvent |= FD_WRITE | FD_CONNECT;
   
   WSAAsyncSelect(info.handle, hWnd_, WM_WINSOCK, lEvent);

   //
   // Add new handler info
   //
   handlerInfoList_.push_front(info);
}

void WConsoleReactor::unregister_handler(EventHandlerBase_ptr handler)
{
   OB_SYNCHRONIZED

   assert(!is_nil(handler));

   for (OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
        iter != handlerInfoList_.end(); iter++)
   {
      HandlerInfo& info = *iter;

      if(info.handler.in() == handler)
      {
         //
         // Handler info found, remove notification messages and
         // invalidate the handler
         //
         WSAAsyncSelect(info.handle, hWnd_, WM_WINSOCK, 0);

         handlerInfoList_.erase(iter);
         return;
      }
   }

   assert(false);
}

bool WConsoleReactor::dispatch(TypeMask typeMask)
{
   interrupt_ = false;
   while(!interrupt_)
   {
      //
      // Exit if a type mask has been set, and if there are no
      // registered event handlers that match this type mask
      //
      TypeMask allTypeMasks = evalTypeMasks();
      if (typeMask != 0 && !(typeMask & allTypeMasks))
         return true;

      MSG msg;
      if (!GetMessage(&msg, 0, 0, 0))
         break;

      //
      // Process one Windows message
      //
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }

   return false;
}

//
// ASYNC SAFE
//
void WConsoleReactor::interrupt_dispatch()
{
   //
   // Don't interrupt twice
   //
   if (interrupt_)
      return;

   //
   // Set the interrupt flag
   //
   interrupt_ = true;

   //
   // Interrupt the Reactor
   //
   ::PostMessage(hWnd_, WM_WINSOCK, 0, 0);
}

//
// True timeouts are not implemented by this reactor, only -1 (block
// forever) and 0 (poll) are handled correctly
//
// TODO: Shouldn't this method translate positive values to -1 (or
// perhaps 0?)
//
bool WConsoleReactor::dispatch_one_event(Long timeout)
{
   while(true)
   {
      MSG msg;
      BOOL result;
      
      if (timeout == -1)
         result = GetMessage(&msg, 0, 0, 0);
      else if (timeout == 0)
         result = PeekMessage(&msg, 0, 0, 0, PM_REMOVE);

      if (result)
      {
         //
         // Process one Windows message
         //
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
      return true;
   }
   return true;
}

bool WConsoleReactor::event_ready()
{
   MSG msg;
   return PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE) ? true : false;
}

LRESULT WConsoleReactor::wndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   if(hWnd == hWnd_ && message == WM_WINSOCK)
   {
      //
      // Use a copy of the handler info, since the handler could
      // change or unregister during event execution.
      //
      HandlerInfo info;

      {
         OB_SYNCHRONIZED

         for (OB_STL::list<HandlerInfo>::iterator iter = handlerInfoList_.begin();
              iter != handlerInfoList_.end(); iter++)
         {

            if ((WPARAM)(*iter).handle == wParam)
               info = *iter;
         }
      }

      if(!is_nil(info.handler))
      {
         //
         // Calculate the event mask
         //
         Mask mask = 0;
         bool loop = false;
         switch(WSAGETSELECTEVENT(lParam))
         {
         case FD_READ:
         case FD_ACCEPT:
         case FD_CLOSE:
            mask |= EventRead;
            break;
      
         case FD_WRITE:
         case FD_CONNECT:
            mask |= EventWrite;
            break;
         }
      
         //
         // Execute the event
         //
         if(info.mask & mask)
         {
            //
            // For write events, we must call handleEvent() in a
            // loop, until writing fails. From the WSAAsyncSelect
            // documentation:
            //
            //  The FD_WRITE event is handled slightly
            //  differently. An FD_WRITE message is posted when a
            //  socket is first connected with connect/WSAConnect
            //  (after FD_CONNECT, if also registered) or accepted
            //  with accept/WSAAccept, and then after a send
            //  operation fails with WSAEWOULDBLOCK and buffer
            //  space becomes available. Therefore, an application
            //  can assume that sends are possible starting from
            //  the first FD_WRITE message and lasting until a
            //  send returns WSAEWOULDBLOCK. After such a failure
            //  the application will be notified that sends are
            //  again possible with an FD_WRITE message.
            //
            Mask retMask;
            do
            {
               retMask = info.handler -> handleEvent(mask);
            }
            while(WSAGETSELECTEVENT(lParam) == FD_WRITE && retMask & EventWrite);
         }
      }

      return 0;
   }
   else if(hWnd == hWnd_ && message == WM_CONSOLE_INPUT)
   {
      if (lParam)
      {
         ThreadState* state = (ThreadState*) lParam;
         lua_dostring(state->code);
         if (state->handle)
            SetEvent(state->handle);
      }
   }

   return DefWindowProc(hWnd, message, wParam, lParam);
}

// ----------------------------------------------------------------------
// Initialization functions
// ----------------------------------------------------------------------

Reactor_ptr WConsoleReactor::Init(const char* title)
{
   HINSTANCE hInstance = GetModuleHandle(NULL);
   SetConsoleTitle(title);
   state_.hwnd = 0;
   state_.handle = 0;
   state_.code[0]= 0;

   return new WConsoleReactor(hInstance);
}

